#!python3
# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Validation pipeline configuration data."""

import enum
import pathlib
import sys

import click

from .console import report_failure, report_success
from .pipeline_config import (
    PipelineConfiguration,
    PipelineConfigurationError,
    PipelineConfigurationPaths,
)

DEFAULT_CONFIGURATION_FILES = {
    "clients": "clients.yml",
    "deployments": "deployments.yml",
    "release_states": "release_states.yml",
    "subscriptions": "subscriptions.yml",
    "systems": "systems.yml",
    "tenants": "tenants.yml",
}


@enum.unique
class ExitState(enum.Enum):
    """Validate configuration script exit states."""

    UNKNOWN = 100
    MISSING_GITREF = 101
    GITREF_PARSE_FAILURE = 102


@click.command()
@click.argument(
    "configuration_dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
)
def _main(configuration_dir: str) -> None:
    """Validate pipeline configuration files.

    Exits non-zero if validation fails.

    CONFIGURATION_DIR is the path to the directory where pipeline configuration
    files are located.
    """
    try:
        this_dir = pathlib.Path(configuration_dir)
        paths = PipelineConfigurationPaths(
            **{x: this_dir / y for x, y in DEFAULT_CONFIGURATION_FILES.items()}
        )
        PipelineConfiguration.from_files(paths)

        report_success("pipeline configuration validated")
    except PipelineConfigurationError as e:
        report_failure(str(e))
        sys.exit(ExitState.MISSING_GITREF.value)
    except Exception as e:
        report_failure("unknown error, {0}".format(str(e)))
        sys.exit(ExitState.UNKNOWN.value)


def flit_entry() -> None:
    """Flit script entry function for ``validate_configuration`` utility."""
    _main()



# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Azure CD release flow utility."""

import enum
import sys

import click

from .._exceptions import GitReferenceError
from ..console import report_failure
from ._cd_release_id import identify_release_id
from ._cd_release_state import identify_release_state
from ._exceptions import ReleaseIdentityError, ReleaseStateError

COMMAND_OUTPUT_TYPES = [
    "id",
    "state",
]


@enum.unique
class ExitState(enum.Enum):
    """Release flow utility CLI exit states."""

    UNKNOWN = 100
    MISSING_GITREF = 101
    GITREF_PARSE_FAILURE = 102


def _do_release_state_actions(git_reference: str) -> None:
    state = identify_release_state(git_reference)
    click.echo("{0}".format(state.name), nl=False)


def _do_release_id_actions(git_reference: str) -> None:
    this_id = identify_release_id(git_reference)
    click.echo("{0}".format(this_id), nl=False)


COMMAND_ACTIONS = {
    "id": _do_release_id_actions,
    "state": _do_release_state_actions,
}


@click.command()
@click.argument(
    "output_type",
    type=click.Choice(COMMAND_OUTPUT_TYPES),
)
@click.argument(
    "git_reference",
    type=str,
)
def azure_subcommand(output_type: str, git_reference: str) -> None:
    """
    From git reference extract Azure deployment release state or release id.

    Report result to stdout.
    """
    try:
        COMMAND_ACTIONS[output_type](git_reference)
    except GitReferenceError as e:
        report_failure(str(e))
        sys.exit(ExitState.MISSING_GITREF)
    except (ReleaseStateError, ReleaseIdentityError) as e:
        report_failure(str(e))
        sys.exit(ExitState.GITREF_PARSE_FAILURE)
    except Exception as e:
        report_failure("unknown error; {0}".format(str(e)))
        sys.exit(ExitState.UNKNOWN)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with foodx_devops_tools.
#  If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.pipeline_config import (
    ClientsDefinitionError,
    load_clients,
)


@pytest.fixture
def apply_clients_test(apply_pipeline_config_test):
    def _apply(mock_content: str):
        result = apply_pipeline_config_test(mock_content, load_clients)
        return result

    return _apply


def test_single_default(apply_clients_test):
    file_text = """---
clients:
  somename:
    release_states:
      - std
      - prd
    system: egms
"""

    result = apply_clients_test(file_text)

    assert len(result.clients) == 1
    assert result.clients["somename"].depends_on is None
    assert result.clients["somename"].pseudonym is None
    assert result.clients["somename"].release_states == ["std", "prd"]
    assert result.clients["somename"].system == "egms"


def test_pseudonym(apply_clients_test):
    file_text = """---
clients:
  somename:
    pseudonym: really long name
    release_states:
      - std
      - prd
    system: egms
"""

    result = apply_clients_test(file_text)

    assert len(result.clients) == 1
    assert result.clients["somename"].pseudonym == "really long name"


def test_depends_on_default(apply_clients_test):
    file_text = """---
clients:
  name1:
    release_states:
      - dev
      - qa
    system: sys1
  name2:
    depends_on: 
      name1:
    release_states:
      - stg
      - prd
    system: sys2
"""

    result = apply_clients_test(file_text)

    assert len(result.clients) == 2
    assert result.clients["name2"].depends_on["name1"] == "qa"


def test_depends_on(apply_clients_test):
    file_text = """---
clients:
  name1:
    release_states:
      - dev
      - qa
      - stg
    system: sys1
  name2:
    depends_on: 
      name1: qa
    release_states:
      - stg
      - prd
    system: sys2
"""

    result = apply_clients_test(file_text)

    assert len(result.clients) == 2
    assert result.clients["name2"].depends_on["name1"] == "qa"


def test_multiple(apply_clients_test):
    file_text = """---
clients:
  name1:
    release_states:
      - dev
      - qa
    system: sys1
  name2:
    release_states:
      - stg
      - prd
    system: sys2
"""

    result = apply_clients_test(file_text)

    assert len(result.clients) == 2
    assert result.clients["name1"].system == "sys1"
    assert result.clients["name2"].system == "sys2"


def test_multiple_depends_on_single(apply_clients_test):
    file_text = """---
clients:
  name1:
    release_states:
      - dev
      - qa
    system: sys1
  name2:
    depends_on: 
      name1:
    release_states:
      - stg
      - prd
    system: sys2
  name3:
    depends_on: 
      name1:
    release_states:
      - stg
      - prd
    system: sys3
"""

    result = apply_clients_test(file_text)

    assert len(result.clients) == 3
    assert result.clients["name2"].depends_on["name1"] == "qa"
    assert result.clients["name3"].depends_on["name1"] == "qa"


def test_missing_dependency_raises(apply_clients_test):
    file_text = """---
clients:
  somename:
    depends_on: 
      othername:
    release_states:
      - stg
      - prd
    system: egms
"""

    with pytest.raises(
        ClientsDefinitionError, match=r"Error validating clients definition"
    ):
        apply_clients_test(file_text)


def test_none_raises(apply_clients_test):
    file_text = """---
"""

    with pytest.raises(
        ClientsDefinitionError, match=r"Error validating clients definition"
    ):
        apply_clients_test(file_text)


def test_empty_raises(apply_clients_test):
    file_text = """---
clients:
"""

    with pytest.raises(
        ClientsDefinitionError, match=r"Error validating clients definition"
    ):
        apply_clients_test(file_text)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy
import pathlib

import pytest

from foodx_devops_tools.pipeline_config import (
    ClientsDefinition,
    DeploymentsDefinition,
    PipelineConfiguration,
    PipelineConfigurationError,
    PipelineConfigurationPaths,
    ReleaseStatesDefinition,
    SubscriptionsDefinition,
    SystemsDefinition,
    TenantsDefinition,
)

MOCK_RESULTS = {
    "clients": ClientsDefinition.parse_obj(
        {
            "clients": {
                "c1": {
                    "release_states": [
                        "r1",
                        "r2",
                    ],
                    "system": "sys1",
                },
                "c2": {
                    "depends_on": {"c1": "r1"},
                    "release_states": [
                        "r2",
                    ],
                    "system": "sys2",
                },
            },
        }
    ),
    "release_states": ReleaseStatesDefinition.parse_obj(
        {"release_states": ["r1", "r2"]}
    ),
    "deployments": DeploymentsDefinition.parse_obj(
        {
            "deployments": {
                "sys1-c1-r1": {
                    "locations": ["l1", "l2"],
                    "subscription": "sub1",
                },
            }
        }
    ),
    "subscriptions": SubscriptionsDefinition.parse_obj(
        {
            "subscriptions": {
                "sub1": {
                    "ado_service_connection": "some-name",
                    "azure_id": "abc123",
                    "tenant": "t1",
                },
            },
        }
    ),
    "systems": SystemsDefinition.parse_obj({"systems": ["sys1", "sys2"]}),
    "tenants": TenantsDefinition.parse_obj(
        {"tenants": {"t1": {"azure_id": "123abc"}}}
    ),
}


@pytest.fixture()
def mock_loads(mocker):
    def _apply(mock_data: dict) -> None:
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_clients",
            return_value=mock_data["clients"],
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_release_states",
            return_value=mock_data["release_states"],
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_deployments",
            return_value=mock_data["deployments"],
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_subscriptions",
            return_value=mock_data["subscriptions"],
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_systems",
            return_value=mock_data["systems"],
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_tenants",
            return_value=mock_data["tenants"],
        )

    return _apply


MOCK_PATHS = PipelineConfigurationPaths(
    clients=pathlib.Path("client/path"),
    release_states=pathlib.Path("release_state/path"),
    deployments=pathlib.Path("deployment/path"),
    subscriptions=pathlib.Path("subscription/path"),
    systems=pathlib.Path("system/path"),
    tenants=pathlib.Path("tenant/path"),
)


class TestPipelineConfiguration:
    def test_default(self, mock_loads):
        mock_loads(MOCK_RESULTS)

        under_test = PipelineConfiguration.from_files(MOCK_PATHS)

        assert len(under_test.clients) == 2
        assert "c1" in under_test.clients
        assert len(under_test.release_states) == 2
        assert under_test.release_states[0] == "r1"
        assert len(under_test.subscriptions) == 1
        assert (
            under_test.subscriptions["sub1"].ado_service_connection
            == "some-name"
        )
        assert under_test.subscriptions["sub1"].azure_id == "abc123"
        assert len(under_test.systems) == 2
        assert under_test.systems[0] == "sys1"

        assert len(under_test.tenants) == 1
        assert under_test.tenants["t1"].azure_id == "123abc"

    def test_bad_deployment_name_raises(self, mock_loads):
        mock_results = copy.deepcopy(MOCK_RESULTS)
        mock_results["deployments"] = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "badname": {
                        "locations": ["l1", "l2"],
                        "subscription": "sub1",
                    },
                }
            }
        )
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad deployment tuple"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_deployment_clients_raises(self, mock_loads):
        mock_results = copy.deepcopy(MOCK_RESULTS)
        mock_results["deployments"] = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "sys1-bad-r1": {
                        "locations": ["l1", "l2"],
                        "subscription": "sub1",
                    },
                }
            }
        )
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad client in deployment tuple"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_deployment_release_states_raises(self, mock_loads):
        mock_results = copy.deepcopy(MOCK_RESULTS)
        mock_results["deployments"] = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "s1-c1-bad": {
                        "locations": ["l1", "l2"],
                        "subscription": "subname",
                    },
                }
            }
        )
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError,
            match=r"Bad release state in deployment tuple",
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_deployment_systems_raises(self, mock_loads):
        mock_results = copy.deepcopy(MOCK_RESULTS)
        mock_results["deployments"] = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "bad-c1-r1": {
                        "locations": ["l1", "l2"],
                        "subscription": "subname",
                    },
                }
            }
        )
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad system in deployment tuple"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_deployment_subscription_raises(self, mock_loads):
        mock_results = copy.deepcopy(MOCK_RESULTS)
        mock_results["deployments"] = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "sys1-c1-r1": {
                        "locations": ["l1", "l2"],
                        "subscription": "bad_name",
                    },
                }
            }
        )
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError,
            match=r"Bad subscription in deployment",
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_client_release_states_raises(self, mock_loads):
        mock_results = copy.deepcopy(MOCK_RESULTS)
        mock_results["clients"] = ClientsDefinition.parse_obj(
            {
                "clients": {
                    "c1": {"release_states": ["bad_state"], "system": "s1"},
                },
            }
        )
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad release state in client"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_client_system_raises(self, mock_loads):
        mock_results = copy.deepcopy(MOCK_RESULTS)
        mock_results["clients"] = ClientsDefinition.parse_obj(
            {
                "clients": {
                    "c1": {"release_states": ["r1"], "system": "bad_system"},
                },
            }
        )
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad system in client"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_subscription_tenant_raises(self, mock_loads):
        mock_results = copy.deepcopy(MOCK_RESULTS)
        mock_results["subscriptions"] = SubscriptionsDefinition.parse_obj(
            {
                "subscriptions": {
                    "sub1": {
                        "ado_service_connection": "some-name",
                        "azure_id": "abc123",
                        "tenant": "bad_name",
                    },
                },
            }
        )
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad tenant in subscription"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with foodx_devops_tools.
#  If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.pipeline_config import (
    DeploymentsDefinitionError,
    load_deployments,
)


@pytest.fixture
def apply_deployments_test(apply_pipeline_config_test):
    def _apply(mock_content: str):
        result = apply_pipeline_config_test(mock_content, load_deployments)
        return result

    return _apply


def test_single(apply_deployments_test):
    file_text = """
---
deployments:
  name:
    locations:
      - loc1
      - loc2
    subscription: some-name
"""

    result = apply_deployments_test(file_text)

    assert len(result.deployments) == 1
    assert "name" in result.deployments
    assert result.deployments["name"].locations == ["loc1", "loc2"]
    assert result.deployments["name"].subscription == "some-name"


def test_multiple(apply_deployments_test):
    file_text = """
---
deployments:
  name1:
    locations:
      - loc1
      - loc2
    subscription: some-name
  name2:
    locations:
      - loc1
      - loc3
    subscription: some-name
  name3:
    locations:
      - loc1
      - loc3
      - loc4
    subscription: other-name
"""

    result = apply_deployments_test(file_text)

    assert len(result.deployments) == 3
    assert "name1" in result.deployments
    assert "name2" in result.deployments
    assert "name3" in result.deployments
    # Assume that name2 data is correct if name1, name3 are correct.
    assert result.deployments["name1"].locations == ["loc1", "loc2"]
    assert result.deployments["name1"].subscription == "some-name"
    assert result.deployments["name3"].locations == ["loc1", "loc3", "loc4"]
    assert result.deployments["name3"].subscription == "other-name"


def test_bad_field_raises(apply_deployments_test):
    file_text = """
---
deployments:
  name:
    bad_field:
      - loc1
      - loc2
    subscription: some-name
"""

    with pytest.raises(
        DeploymentsDefinitionError,
        match=r"Error validating deployments definition",
    ):
        apply_deployments_test(file_text)


def test_none_raises(apply_deployments_test):
    file_text = """
---
"""

    with pytest.raises(
        DeploymentsDefinitionError,
        match=r"Error validating deployments definition",
    ):
        apply_deployments_test(file_text)


def test_empty_list_raises(apply_deployments_test):
    file_text = """
---
deployments:
"""

    with pytest.raises(
        DeploymentsDefinitionError,
        match=r"Error validating deployments definition",
    ):
        apply_deployments_test(file_text)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Release states deployment configuration I/O."""

import pathlib
import typing

import pydantic

from ._loader import load_configuration

ENTITY_NAME = "release_states"


class ReleaseStatesDefinitionError(Exception):
    """Problem loading release state definitions."""


T = typing.TypeVar("T", bound="ReleaseStatesDefinition")

ValueType = typing.List[str]


class ReleaseStatesDefinition(pydantic.BaseModel):
    """Definition of deployment release states."""

    release_states: ValueType

    @pydantic.validator(ENTITY_NAME)
    def check_release_states(
        cls: pydantic.BaseModel, value: ValueType
    ) -> ValueType:
        """Validate ``release_states`` field."""
        if not value:
            raise ValueError("Empty release state names prohibited")
        if len(set(value)) != len(value):
            raise ValueError("Duplicate release state names prohibited")

        return value


def load_release_states(file_path: pathlib.Path) -> ReleaseStatesDefinition:
    """
    Load release state definitions from file.

    Args:
        file_path: Path to system definitions file.

    Returns:
        Release state definitions.
    Raises:
        ReleaseStateDefinitionError: If an error occurs loading the file.
    """
    result = load_configuration(
        file_path,
        ReleaseStatesDefinition,
        ReleaseStatesDefinitionError,
        ENTITY_NAME,
    )

    return typing.cast(ReleaseStatesDefinition, result)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib
import typing

import pydantic
import pytest


@pytest.fixture
def apply_pipeline_config_test(mocker):
    def _apply(
        mock_content: str,
        method_under_test: typing.Callable[[pathlib.Path], pydantic.BaseModel],
    ):
        mock_path = mocker.create_autospec(pathlib.Path)("some/path")
        mock_path.open = mocker.mock_open(read_data=mock_content)

        result = method_under_test(mock_path)

        return result

    return _apply



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib
import typing

import pydantic
import ruamel.yaml  # type: ignore # noqa: F401
from ruamel.yaml import YAML  # type: ignore


def load_configuration(
    file_path: pathlib.Path,
    data_model: typing.Type[pydantic.BaseModel],
    error_type: typing.Type[Exception],
    entity_name: str,
) -> pydantic.BaseModel:
    """
    Load definitions from file.

    Args:
        file_path: Path to definitions file.
        data_model: Pydantic data model to use for the data.
        error_type: Exception type to use for reporting errors.
        entity_name: Name of the entity being loaded.

    Returns:
        ``data_model`` object with populated data.
    Raises:
        error_type: If an error occurs loading the file.
    """
    with file_path.open(mode="r") as f:
        loader = YAML(typ="safe")
        yaml_data = loader.load(f)

    try:
        result = data_model.parse_obj(yaml_data)

        return result
    except pydantic.ValidationError as e:
        raise error_type(
            "Error validating {0} definition".format(
                entity_name.replace("_", " ")
            )
        ) from e



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Manage pipeline configuration files collectively."""

import dataclasses
import pathlib
import re
import typing

import pydantic

from .clients import ValueType as ClientsData
from .clients import load_clients
from .deployments import ValueType as DeploymentsData
from .deployments import load_deployments
from .release_states import ValueType as ReleaseStatesData
from .release_states import load_release_states
from .subscriptions import ValueType as SubscriptionsData
from .subscriptions import load_subscriptions
from .systems import ValueType as SystemsData
from .systems import load_systems
from .tenants import ValueType as TenantsData
from .tenants import load_tenants

DEPLOYMENT_NAME_REGEX = (
    r"^(?P<system>[a-z0-9]+)"
    r"[_-]"
    r"(?P<client>[a-z0-9]+)"
    r"[_-]"
    r"(?P<release_state>[a-z0-9]+)$"
)


class PipelineConfigurationError(Exception):
    """Problem occurred loading pipeline configuration."""


@dataclasses.dataclass()
class PipelineConfigurationPaths:
    """Paths to pipeline configuration files."""

    clients: pathlib.Path
    release_states: pathlib.Path
    deployments: pathlib.Path
    subscriptions: pathlib.Path
    systems: pathlib.Path
    tenants: pathlib.Path


T = typing.TypeVar("T", bound="PipelineConfiguration")


class PipelineConfiguration(pydantic.BaseModel):
    """Pipeline configuration data."""

    clients: ClientsData
    release_states: ReleaseStatesData
    deployments: DeploymentsData
    subscriptions: SubscriptionsData
    systems: SystemsData
    tenants: TenantsData

    @classmethod
    def from_files(cls: typing.Type[T], paths: PipelineConfigurationPaths) -> T:
        """
        Load pipeline configuration from collection of files.

        Args:
            paths: Paths to pipeline configuration files.

        Returns:
            Instantiated ``PipelineConfiguration`` object.
        """
        client_config = load_clients(paths.clients)
        release_state_config = load_release_states(paths.release_states)
        deployment_config = load_deployments(paths.deployments)
        subscription_config = load_subscriptions(paths.subscriptions)
        system_config = load_systems(paths.systems)
        tenant_config = load_tenants(paths.tenants)
        kwargs = {
            "clients": client_config.clients,
            "release_states": release_state_config.release_states,
            "deployments": deployment_config.deployments,
            "subscriptions": subscription_config.subscriptions,
            "systems": system_config.systems,
            "tenants": tenant_config.tenants,
        }
        cls._validate_clients(kwargs)
        cls._validate_deployments(kwargs)
        cls._validate_subscriptions(kwargs)
        new_object = cls(**kwargs)

        return new_object

    @staticmethod
    def _validate_clients(loaded_data: dict) -> None:
        for name, data in loaded_data["clients"].items():
            if not all(
                [
                    x in loaded_data["release_states"]
                    for x in data.release_states
                ]
            ):
                raise PipelineConfigurationError(
                    "Bad release state in client, {0}".format(name)
                )

            if data.system not in loaded_data["systems"]:
                raise PipelineConfigurationError(
                    "Bad system in client, {0}".format(name)
                )

    @staticmethod
    def _validate_deployments(loaded_data: dict) -> None:
        """
        Validate loaded deployment data.

        Args:
            loaded_data: Data loaded from pipeline configuration.
        """
        deployment_names = set(loaded_data["deployments"].keys())
        this_re = re.compile(DEPLOYMENT_NAME_REGEX)
        for this_name in deployment_names:
            result = this_re.match(this_name)

            if not result:
                raise PipelineConfigurationError(
                    "Bad deployment tuple, {0}".format(this_name)
                )

            this_client = result.group("client")
            if this_client not in loaded_data["clients"]:
                raise PipelineConfigurationError(
                    "Bad client in deployment tuple, {0}".format(this_client)
                )

            this_release_state = result.group("release_state")
            if this_release_state not in loaded_data["release_states"]:
                raise PipelineConfigurationError(
                    "Bad release state in deployment tuple, {0}".format(
                        this_release_state
                    )
                )

            this_system = result.group("system")
            if this_system not in loaded_data["systems"]:
                raise PipelineConfigurationError(
                    "Bad system in deployment tuple, {0}".format(this_system)
                )

            this_subscription = loaded_data["deployments"][
                this_name
            ].subscription
            if this_subscription not in loaded_data["subscriptions"]:
                raise PipelineConfigurationError(
                    "Bad subscription in deployment, {0}".format(this_name)
                )

    @staticmethod
    def _validate_subscriptions(loaded_data: dict) -> None:
        for name, data in loaded_data["subscriptions"].items():
            if data.tenant not in loaded_data["tenants"]:
                raise PipelineConfigurationError(
                    "Bad tenant in subscription, {0}".format(name)
                )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import os
import pathlib

import pytest
from click.testing import CliRunner


@pytest.fixture()
def click_runner() -> CliRunner:
    return CliRunner()


@contextlib.contextmanager
def isolated_filesystem(expected_dir, mock_runner):
    with mock_runner.isolated_filesystem():
        os.makedirs(expected_dir)

        yield pathlib.Path(expected_dir).absolute()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import json
import os
import pathlib
import typing
import uuid

import aiofiles
import pytest
from ruamel.yaml import YAML

from foodx_devops_tools.puff.arm import (
    ArmTemplateError,
    _delete_parameter_file,
    _do_file_actions,
    _linearize_environments,
    _linearize_name,
    _linearize_regions,
    _linearize_services,
    _remove_keys,
    _save_parameter_file,
    do_arm_template_parameter_action,
    load_yaml,
)


def initialize_filesystem(file_name: pathlib.Path, file_content: str):
    with file_name.open("w") as f:
        f.write(file_content)


class TestLinearizeName:
    MOCK_BASE = {
        "environments": dict(),
        "services": dict(),
    }

    def test_level0_name(self):
        expected_result = {
            "another-name": {
                "environments": dict(),
                "services": dict(),
            },
        }
        mock_file = "some-file"
        mock_input = {
            **self.MOCK_BASE,
            "name": "another-name",
        }

        result = _linearize_name(mock_input, mock_file)

        assert result == expected_result
        assert "name" not in result["another-name"]

    def test_level0_noname(self):
        expected_result = {
            "some-file": {
                "environments": dict(),
                "services": dict(),
            },
        }
        mock_file = "some-file"
        mock_input = self.MOCK_BASE.copy()

        result = _linearize_name(mock_input, mock_file)

        assert result == expected_result
        assert "name" not in result["some-file"]


class TestLinearizedServices:
    def test_services_environments(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "environments": {
                    "e1": {
                        "k1": "be1k1",
                        "k2": "be1k2",
                    },
                },
                "services": {
                    "s1": {
                        "p2": "s1p2",
                        "environments": {
                            "e1": {
                                "k1": "s1e1k1",
                            },
                        },
                    },
                },
            },
        }
        expected_result = {
            "some-file.s1": {
                "p1": "bp1",
                "p2": "s1p2",
                "environments": {
                    "e1": {
                        "k1": "s1e1k1",
                        "k2": "be1k2",
                    },
                },
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_multiple_services(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "services": {
                    "s1": {
                        "p2": "s1p2",
                    },
                    "s2": {
                        "p2": "s2p2",
                    },
                },
            },
        }
        expected_result = {
            "some-file.s1": {
                "p1": "bp1",
                "p2": "s1p2",
            },
            "some-file.s2": {
                "p1": "bp1",
                "p2": "s2p2",
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_services_environments(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "environments": {
                    "e1": {
                        "k1": "be1k1",
                        "k2": "be1k2",
                    },
                },
                "services": {
                    "s1": {
                        "p2": "s1p2",
                        "environments": {
                            "e1": {
                                "k1": "s1e1k1",
                            },
                        },
                    },
                },
            },
        }
        expected_result = {
            "some-file.s1": {
                "p1": "bp1",
                "p2": "s1p2",
                "environments": {
                    "e1": {
                        "k1": "s1e1k1",
                        "k2": "be1k2",
                    },
                },
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_empty_services(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "environments": {
                    "e1": {
                        "k1": "be1k1",
                        "k2": "be1k2",
                    },
                },
                "services": dict(),
            },
        }
        expected_result = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "environments": {
                    "e1": {
                        "k1": "be1k1",
                        "k2": "be1k2",
                    },
                },
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_no_services_no_environments(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "services": dict(),
            },
        }
        expected_result = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_no_services_no_environments_empty_base(self):
        mock_base = {
            "some-file": dict(),
        }
        expected_result = mock_base
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_services_empty_base(self):
        mock_base = {
            "some-file": {
                "services": {
                    "s1": {
                        "p2": "s1p2",
                        "environments": {
                            "e1": {
                                "k1": "s1e1k1",
                            },
                        },
                    },
                },
            },
        }
        expected_result = {
            "some-file.s1": {
                "p2": "s1p2",
                "environments": {
                    "e1": {
                        "k1": "s1e1k1",
                    },
                },
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result


class TestLinearizedEnvironments:
    def test_environments_regions(self):
        mock_base = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
                "environments": {
                    "e1": {
                        "k1": "e1k1",
                        "regions": [
                            {"r1": {"k1": "e1r1k1"}},
                            {"r2": {"k2": "e1r2k2"}},
                        ],
                    },
                },
            },
        }
        expected_result = {
            "this.stub.e1": {
                "p1": "bp1",
                "k1": "e1k1",
                "k2": "bk2",
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_multiple_stubs_multiple_environments(self):
        mock_base = {
            "stub1": {
                "p1": "stub1.p1",
                "k2": "stub1.k2",
                # "full" environment
                "environments": {
                    "e1": {
                        "k1": "e1k1",
                    },
                    "e2": {
                        "k1": "e2k1",
                    },
                },
            },
        }
        expected_result = {
            "stub1.e1": {
                "p1": "stub1.p1",
                "k1": "e1k1",
                "k2": "stub1.k2",
            },
            "stub1.e2": {
                "p1": "stub1.p1",
                "k1": "e2k1",
                "k2": "stub1.k2",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_multiple_stubs_mixed_environments(self):
        mock_base = {
            "stub1": {
                "p1": "stub1.p1",
                "k2": "stub1.k2",
                # "full" environment
                "environments": {
                    "e1": {
                        "k1": "e1k1",
                        "regions": [
                            {"r1": {"k1": "e1r1k1"}},
                            {"r2": {"k2": "e1r2k2"}},
                        ],
                    },
                },
            },
            "stub2": {
                "p1": "stub2.p1",
                "k2": "stub2.k2",
                # empty environment
                "environments": dict(),
            },
        }
        expected_result = {
            "stub1.e1": {
                "p1": "stub1.p1",
                "k1": "e1k1",
                "k2": "stub1.k2",
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
            "stub2": {
                "p1": "stub2.p1",
                "k2": "stub2.k2",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_empty_environments(self):
        mock_base = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
                "environments": dict(),
            },
        }
        expected_result = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_environments_no_environment_regions(self):
        mock_base = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
                "environments": {
                    "e1": {
                        "k1": "e1k1",
                    },
                },
            },
        }
        expected_result = {
            "this.stub.e1": {
                "p1": "bp1",
                "k1": "e1k1",
                "k2": "bk2",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_no_environments_no_regions(self):
        mock_base = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
            },
        }
        expected_result = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_no_environments_no_regions_empty_base(self):
        mock_base = {
            "this.stub": dict(),
        }
        expected_result = {
            "this.stub": dict(),
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_environments_empty_base(self):
        mock_base = {
            "this.stub": {
                "environments": {
                    "e1": {
                        "k1": "e1k1",
                        "regions": [
                            {"r1": {"k1": "e1r1k1"}},
                            {"r2": {"k2": "e1r2k2"}},
                        ],
                    },
                },
            },
        }
        expected_result = {
            "this.stub.e1": {
                "k1": "e1k1",
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result


class TestLinearizedRegions:
    def test_regions(self):
        mock_base = {
            "this.stub": {
                "k1": "bk1",
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
        }
        expected_result = {
            "this.stub.r1": {
                "k1": "e1r1k1",
            },
            "this.stub.r2": {
                "k1": "bk1",
                "k2": "e1r2k2",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_multiple_stubs_mixed_regions(self):
        mock_base = {
            "stub1": {
                "k1": "stub1.k1",
                # "full" region in a stub
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
            "stub2": {
                "k1": "stub2.k1",
                # empty regions in a stub
                "regions": list(),
            },
        }
        expected_result = {
            "stub1.r1": {
                "k1": "e1r1k1",
            },
            "stub1.r2": {
                "k1": "stub1.k1",
                "k2": "e1r2k2",
            },
            "stub2": {
                "k1": "stub2.k1",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_empty_regions(self):
        mock_base = {
            "this.stub": {
                "k1": "bk1",
                "regions": list(),
            },
        }
        expected_result = {
            "this.stub": {
                "k1": "bk1",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_no_regions(self):
        mock_base = {
            "this.stub": {
                "k1": "bk1",
            },
        }
        expected_result = {
            "this.stub": {
                "k1": "bk1",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_regions_empty_base(self):
        mock_base = {
            "this.stub": {
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
        }
        expected_result = {
            "this.stub.r1": {
                "k1": "e1r1k1",
            },
            "this.stub.r2": {
                "k2": "e1r2k2",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_no_regions_empty_base(self):
        mock_base = {
            "this.stub": dict(),
        }
        expected_result = {
            "this.stub": dict(),
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result


class TestRemoveKeys:
    def test_clean(self):
        data = {
            "name": "some name",
            "p1": "v1",
            "p2": 3.14159,
            "environments": dict(),
            "services": dict(),
        }
        pre_copy = data.copy()
        result = _remove_keys(data, ["name", "environments", "services"])

        assert result == {
            "p1": "v1",
            "p2": 3.14159,
        }
        # original data must not be modified
        assert data == pre_copy


class TestSaveParameterFile:
    @pytest.mark.asyncio
    async def test_create_clean(self, tmp_path_factory):
        data = {
            "p1": "v1",
            "p2": "v2",
        }
        content = await self._do_create_test(data, tmp_path_factory)
        assert "p1" in content["parameters"]
        assert "p2" in content["parameters"]
        assert content["parameters"]["p1"]["value"] == "v1"
        assert content["parameters"]["p2"]["value"] == "v2"

    def _check_basic_content(self, this_file: pathlib.Path) -> dict:
        assert this_file.exists()
        with this_file.open("r") as f:
            content = json.load(f)

        assert all(
            [x in content for x in ["$schema", "contentVersion", "parameters"]]
        )
        return content

    async def _do_create_test(
        self, parameter_data: typing.Optional[dict], path_factory
    ) -> dict:
        this_directory = path_factory.mktemp(str(uuid.uuid4()))
        this_file = this_directory / "some_file"
        assert not this_file.exists()

        await _save_parameter_file(this_file, parameter_data)

        return self._check_basic_content(this_file)

    @pytest.mark.asyncio
    async def test_none_parameters(self, tmp_path_factory):
        await self._do_create_test(None, tmp_path_factory)

    @pytest.mark.asyncio
    async def test_empty_parameters(self, tmp_path_factory):
        await self._do_create_test(dict(), tmp_path_factory)


class TestDeleteParameterFile:
    @pytest.mark.asyncio
    async def test_delete_clean(self, tmp_path_factory):
        this_directory = tmp_path_factory.mktemp(str(uuid.uuid4()))
        this_file = this_directory / "some_file"
        initialize_filesystem(this_file, "some content")

        assert this_file.exists()

        await _delete_parameter_file(this_file)

        assert not this_file.exists()


class TestDoFileActions:
    MOCK_PARAMETER_DATA = {
        "some-file": {
            "k1": "v1",
        }
    }

    @pytest.mark.asyncio
    async def test_delete_clean(self, mocker):
        mock_delete = mocker.patch(
            "foodx_devops_tools.puff.arm._delete_parameter_file"
        )
        mocker.patch("foodx_devops_tools.puff.arm._save_parameter_file")

        await _do_file_actions(
            True, pathlib.Path("some/path"), self.MOCK_PARAMETER_DATA
        )

        mock_delete.assert_called_once()

    @pytest.mark.asyncio
    async def test_create_clean(self, mocker):
        mocker.patch("foodx_devops_tools.puff.arm._delete_parameter_file")
        mock_save = mocker.patch(
            "foodx_devops_tools.puff.arm._save_parameter_file"
        )

        await _do_file_actions(
            False, pathlib.Path("some/path"), self.MOCK_PARAMETER_DATA
        )

        mock_save.assert_called_once()


class TestLoadYaml:
    @pytest.mark.asyncio
    async def test_clean(self, tmp_path_factory):
        this_directory = tmp_path_factory.mktemp("tmp")
        this_path = this_directory / "some.file"
        with this_path.open(mode="w") as f:
            f.write(
                """---
some:
- d1
- d2
"""
            )
        result = await load_yaml(this_path)

        assert result == {"some": ["d1", "d2"]}

    @pytest.mark.asyncio
    async def test_empty_yaml_warns(self, caplog, tmp_path_factory):
        this_directory = tmp_path_factory.mktemp("tmp")
        this_path = this_directory / "some.file"
        with this_path.open(mode="w") as f:
            f.write(
                """---
"""
            )

        result = await load_yaml(this_path)

        assert any(["Empty YAML file" in x for x in caplog.messages])

    @pytest.mark.asyncio
    async def test_missing_file_raises(self, caplog, tmp_path_factory):
        this_directory = tmp_path_factory.mktemp("tmp")
        this_path = this_directory / "some.file"

        with pytest.raises(
            FileNotFoundError, match=r"No such file or " r"directory"
        ):
            await load_yaml(this_path)

        assert any(["File not found" in x for x in caplog.messages])

    @pytest.mark.asyncio
    async def test_directory_raises(self, caplog, tmp_path_factory):
        """Path pointing to a directory raises an exception."""
        this_directory = tmp_path_factory.mktemp("tmp")
        this_path = this_directory / "some.dir"
        os.makedirs(this_path, exist_ok=True)

        with pytest.raises(IsADirectoryError, match=r"Is a directory"):
            await load_yaml(this_path)

        assert any(["Path is a directory" in x for x in caplog.messages])

    @pytest.mark.asyncio
    async def test_unexpected_logs(self, caplog, mocker, tmp_path_factory):
        """Unexpected error logs a report."""
        this_directory = tmp_path_factory.mktemp("tmp")
        mocker.patch(
            "foodx_devops_tools.puff.arm.ruamel.yaml.YAML",
            side_effect=RuntimeError("some error"),
        )
        this_path = this_directory / "some.dir"
        os.makedirs(this_path, exist_ok=True)

        with pytest.raises(RuntimeError, match=r"^some error"):
            await load_yaml(this_path)

        assert any(["Unexpected error with path" in x for x in caplog.messages])


class TestGenerateArmTemplateParameters:
    @pytest.mark.asyncio
    async def test_name_optional(self, mocker):
        puff_content = """---
environments:
  e1:
    k1: e1k1
    k2: e1k2
  e2:
    k1: e2k1
    k2: e2k2
"""
        expected_files = [
            "this_file.e1.json",
            "this_file.e2.json",
        ]
        expected_content = [
            {
                "k1": "e1k1",
                "k2": "e1k2",
            },
            {
                "k1": "e2k1",
                "k2": "e2k2",
            },
        ]

        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    async def _do_file_test(
        self,
        expected_files: typing.List[str],
        expected_values: typing.List[dict],
        puff_content: str,
        this_mocker,
        is_delete_action: bool = False,
    ) -> None:
        this_yaml = YAML(typ="safe")
        data = this_yaml.load(puff_content)
        this_mocker.patch(
            "foodx_devops_tools.puff.arm.load_yaml", return_value=data
        )
        mock_save = this_mocker.patch(
            "foodx_devops_tools.puff.arm._save_parameter_file"
        )
        this_directory = pathlib.Path("some/path")
        puff_path = this_directory / "this_file.yml"

        await do_arm_template_parameter_action(puff_path, is_delete_action)

        expected_paths = [this_directory / x for x in expected_files]

        mock_save.has_calls(
            [
                this_mocker.call(expected_paths[x], expected_values[x])
                for x in range(0, len(expected_paths))
            ]
        )

    @pytest.mark.asyncio
    async def test_environments_regions_clean(self, mocker):
        puff_content = """---
name: some-name
environments:
  e1:
    k1: e1k1
    k2: e1k2
  e2:
    k1: e2k1
    k2: e2k2
    regions:
      - r1:
          k3: e2r1k3
"""
        expected_files = [
            "some-name.e1.json",
            "some-name.e2.r1.json",
        ]
        expected_content = [
            {
                "k1": "e1k1",
                "k2": "e1k2",
            },
            {
                "k1": "e2k1",
                "k2": "e2k2",
                "k3": "e2r1k3",
            },
        ]
        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_services_clean(self, mocker):
        puff_content = """---
name: some-name
services:
  s1:
    k1: s1k1
    k2: s1k2
  s2:
    k1: s2k1
    k2: s2k2
    k3: s2k3
"""
        expected_files = [
            (".".join(["some-name", x, "json"])) for x in ["s1", "s2"]
        ]
        expected_content = [
            {
                "k1": "s1k1",
                "k2": "s1k2",
            },
            {
                "k1": "s2k1",
                "k2": "s2k2",
                "k3": "s2k3",
            },
        ]
        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_environments_clean(self, mocker):
        puff_content = """---
name: some-name
environments:
  e1:
    k1: e1k1
    k2: e1k2
  e2:
    k1: e2k1
    k2: e2k2
    k3: e2k3
"""
        expected_files = [
            (".".join(["some-name", x, "json"])) for x in ["e1", "e2"]
        ]
        expected_content = [
            {
                "k1": "e1k1",
                "k2": "e1k2",
            },
            {
                "k1": "e2k1",
                "k2": "e2k2",
                "k3": "e2k3",
            },
        ]
        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_regions_clean(self, mocker):
        puff_content = """---
name: some-name
environments:
  e1:
    regions:
      - r1:
          k1: e1r1k1
          k2: e1r1k2
      - r2:
          k1: e1r2k1
          k2: e1r2k2
          k3: e1r2k3
"""
        expected_files = [
            (".".join(["some-name", x, "json"])) for x in ["e1.r1", "e1.r2"]
        ]
        expected_content = [
            {
                "k1": "e1r1k1",
                "k2": "e1r1k2",
            },
            {
                "k1": "e1r2k1",
                "k2": "e1r2k2",
                "k3": "e1r2k3",
            },
        ]
        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_all_in_clean(self, mocker):
        puff_content = """---
name: some-name
k1: bk1
k2: bk2
environments:
  e1:
    k3: e1k3
  e2:
    k3: e2k3
    regions:
      - r2:
          k1: e2r2k1
services:
  s1:
    k1: s1k1
    k2: s1k2
    environments:
      e1:
        k5: s1e1k5
        regions:
          - r1:
              k2: s1e1r1k2
              k6: s1e1r1k6
  s2:
    k1: s2k1
    k2: s2k2
    k3: s2k3
"""
        expected_files = [
            (".".join(["some-name", x, "json"]))
            for x in ["s1.e1.r1", "s1.e2.r2", "s2.e1", "s2.e2.r2"]
        ]
        expected_content = [
            {
                "k1": "value1",
                "k2": "s1e1r1k2",
                "k3": "e1k3",
                "k5": "s1e1k5",
                "k6": "s1e1r1k6",
            },
            {
                "k1": "e2r2k1",
                "k2": "bk2",
                "k3": "e2k3",
            },
            {
                "k1": "s2k1",
                "k2": "s2k2",
                "k3": "e1k3",
            },
            {
                "k1": "e2r2k1",
                "k2": "s2k2",
                "k3": "e2k3",
            },
        ]
        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_empty_raises(self, tmp_path_factory):
        puff_content = ""
        this_directory = tmp_path_factory.mktemp(str(uuid.uuid4())).absolute()
        puff_path = this_directory / "this_file.yml"
        initialize_filesystem(puff_path, puff_content)

        with pytest.raises(ArmTemplateError):
            await do_arm_template_parameter_action(puff_path, False)

    @pytest.mark.asyncio
    async def test_empty_services_raises(self, tmp_path_factory):
        puff_content = """---
environments:
  e1:
    v: v
services:
"""
        this_directory = tmp_path_factory.mktemp(str(uuid.uuid4())).absolute()
        puff_path = this_directory / "this_file.yml"
        initialize_filesystem(puff_path, puff_content)

        with pytest.raises(ArmTemplateError):
            await do_arm_template_parameter_action(puff_path, False)



#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#
# https://gitlab.com/ci-cd-devops/build_harness/-/blob/main/tests/ci/unit_tests/test_utility.py

from foodx_devops_tools.utility import run_command


class TestRunCommand:
    def test_simple(self, mocker):
        mock_run = mocker.patch("foodx_devops_tools.utility.subprocess.run")
        command = ["something", "--option"]

        result = run_command(command)

        mock_run.assert_called_once_with(command)
        assert result == mock_run.return_value



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Azure ARM template parameter generation."""

import json
import logging
import pathlib
import typing

import aiofiles  # type: ignore
import aiofiles.os  # type: ignore
import pydantic
import ruamel.yaml
from deepmerge import always_merger  # type: ignore

from ._header import ARMTEMPLATE_PARAMETERS_HEADER
from ._puff_parameters import PuffParameterModel

log = logging.getLogger(__name__)

YamlData = typing.Dict[str, typing.Any]

BASE_PARAMETER_EXCLUDE_LABELS = ["default", "environments", "name", "services"]


class ArmTemplateError(Exception):
    """Problem occurred with ARM template generation."""


async def load_yaml(path: pathlib.Path) -> YamlData:
    """
    Asynchronously load yaml file data.

    Args:
        path: Path to YAML file to load.

    Returns:
        Loaded YAML data.
    Raises:
        FileNotFoundError: If file cannot be loaded.
        IsADirectoryError: If path is a directory instead of a file.
    """
    try:
        yaml = ruamel.yaml.YAML(typ="safe")
        async with aiofiles.open(str(path), mode="r") as f:
            content = await f.read()
            yaml_data = yaml.load(content)

        if not yaml_data:
            log.warning("Empty YAML file, {0}".format(str(path)))

        return yaml_data
    except FileNotFoundError:
        log.error("File not found, {0}".format(str(path)))
        raise
    except IsADirectoryError:
        log.error("Path is a directory, {0}".format(str(path)))
        raise
    except Exception:
        log.error("Unexpected error with path, {0}".format(str(path)))
        raise


def _remove_keys(
    parameters: typing.Dict[str, typing.Any], exclude_labels: typing.List[str]
) -> dict:
    """
    Remove keys from a dictionary without changing the original.

    Attempts to remove keys that don't exist in the dictinary are silently
    ignored.

    Args:
        parameters: Dictionary to be adjusted.

    Returns:
        Modified dictionary.
    """
    this_copy = parameters.copy()
    for k in exclude_labels:
        try:
            del this_copy[k]
        except KeyError:
            # silently ignore missing keys by suppressing this exception
            pass

    return this_copy


async def _save_parameter_file(
    target_path: pathlib.Path,
    parameter_data: typing.Optional[dict],
) -> None:
    """
    Create ARM template parameter file.

    Args:
        target_path: File path to be created or deleted.
        parameter_data: Parameters to be applied to data.
    """
    generated_parameters = ARMTEMPLATE_PARAMETERS_HEADER.copy()
    if parameter_data:
        for item_key, value in parameter_data.items():
            generated_parameters["parameters"][item_key] = {
                "value": value,
            }
    async with aiofiles.open(target_path, mode="w") as f:
        await f.write(json.dumps(generated_parameters))


async def _delete_parameter_file(
    target_path: pathlib.Path,
) -> None:
    """
    Delete generated ARM template parameter file.

    Args:
        target_path: File path to be created or deleted.
    """
    if target_path.exists():
        await aiofiles.os.remove(target_path)


async def _do_file_actions(
    is_delete_action: bool, target_path: pathlib.Path, parameter_data: dict
) -> None:
    """
    Create or delete generated ARM template parameter files.

    Args:
        is_delete_action: Signal deleting or creating parameter files.
        target_path: Directory to store or delete parameter files.
        parameter_data: Data for each ARM template parameter file.
    """
    for key, values in parameter_data.items():
        this_path = target_path / ".".join([key, "json"])
        if is_delete_action:
            await _delete_parameter_file(this_path)
        else:
            await _save_parameter_file(this_path, values)


def _linearize_name(base_data: dict, filename: str) -> dict:
    cleaned = _remove_keys(base_data, ["name"])
    this_name = base_data.get("name", filename)
    if not this_name:
        raise RuntimeError("Invalid empty name")

    linearized = {
        this_name: cleaned,
    }
    return linearized


def _linearize_iterable(base_data: dict, plural: str, singular: str) -> dict:
    linearized = dict()
    for base_key, base_value in base_data.items():
        this_data = (
            base_value[plural].copy() if plural in base_value else dict()
        )
        cleaned = _remove_keys(base_value, [plural])
        if this_data:
            for iterable_key, iterable_data in this_data.items():
                if not iterable_key:
                    raise RuntimeError(
                        "Invalid empty {0} name".format(singular)
                    )

                linearized[
                    "{0}.{1}".format(base_key, iterable_key)
                ] = always_merger.merge(cleaned.copy(), iterable_data)
        else:
            linearized[base_key] = cleaned

    return linearized if linearized else base_data.copy()


def _linearize_services(base_data: dict) -> dict:
    plural = "services"
    singular = "service"
    return _linearize_iterable(base_data, plural, singular)


def _linearize_environments(base_data: dict) -> dict:
    plural = "environments"
    singular = "environment"
    return _linearize_iterable(base_data, plural, singular)


def _linearize_regions(base_data: dict) -> dict:
    plural = "regions"
    singular = "region"
    linearized = dict()
    for base_key, base_value in base_data.items():
        this_data = (
            base_value[plural].copy() if plural in base_value else list()
        )
        cleaned = _remove_keys(base_value, [plural])
        if this_data:
            for this_item in this_data:
                for iterable_key, iterable_data in this_item.items():
                    if not iterable_key:
                        raise RuntimeError(
                            "Invalid empty {0} name".format(singular)
                        )

                    linearized[
                        "{0}.{1}".format(base_key, iterable_key)
                    ] = always_merger.merge(cleaned.copy(), iterable_data)
        else:
            linearized[base_key] = cleaned

    return linearized if linearized else base_data.copy()


def _linearize_parameters(yaml_data: dict, file_name: str) -> dict:
    level0_merge = _linearize_name(yaml_data, file_name)
    level1_merge = _linearize_services(level0_merge)
    level2_merge = _linearize_environments(level1_merge)
    level3_merge = _linearize_regions(level2_merge)

    return level3_merge


async def do_arm_template_parameter_action(
    puff_file_path: pathlib.Path,
    is_delete_action: bool,
) -> None:
    """
    Generate ARM template parameter files from puff YAML file.

    The generated files are always output to the same directory as the source
    puff file.

    Args:
        puff_file_path: Path to source YAML parameter file.
        is_delete_action: True if files should be deleted instead of created.
    """
    target_path = puff_file_path.parent

    yaml_data = await load_yaml(puff_file_path)
    try:
        # for now, just use the pydantic model to validate the YAML data.
        PuffParameterModel.parse_obj(yaml_data)
    except pydantic.ValidationError as e:
        log.error(str(e))
        raise ArmTemplateError("Puff parameter validation failed") from e

    file_name = pathlib.Path(puff_file_path.stem).name
    merged_data = _linearize_parameters(yaml_data, file_name)
    await _do_file_actions(is_delete_action, target_path, merged_data)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#
# https://gitlab.com/ci-cd-devops/build_harness/-/blob/main/build_harness/_utility.py

"""General support for build harness implementation."""

import logging
import subprocess
import typing

log = logging.getLogger(__name__)

CommandArgs = typing.List[str]


def run_command(
    command: CommandArgs, **kwargs: typing.Any
) -> subprocess.CompletedProcess:
    """
    Run a system command using ``subprocess.run``.

    Args:
        command: List of command and arguments.
        **kwargs: Optional arguments passed through to ``subprocess.run``.

    Returns:
        Subprocess results.
    """
    log.debug("command to run, {0}".format(str(command)))
    log.debug("command arguments, {0}".format(str(kwargs)))
    result = subprocess.run(command, **kwargs)

    return result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Declarations of constants for pipeline refereance parsing."""

BRANCH_PREFIX = r"refs/heads/"
TAG_PREFIX = r"refs/tags/"

SEMANTIC_VERSION_GITREF = r"(\d+\.\d+\.\d+)"
